/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

    filter.c

Abstract: NULL filter driver -- boilerplate code

Author:

    ervinp

Environment:

    Kernel mode

Revision History:


--*/

#define INITGUID				// define GUID_CEC_CTG


#include <WDM.H>
#include <ntddser.H>

#include "serial.h" // an excerpt from \src\kernel\serial\serial.h


#if DBG
	// allow symbolic request viewing
	#define IRPMJFUNCDESC
	#include <ksdebug.h>
#endif

#include "filter.h"

#ifdef ALLOC_PRAGMA
        #pragma alloc_text(INIT, DriverEntry)
        #pragma alloc_text(PAGE, Ctg_AddDevice)
        #pragma alloc_text(PAGE, Ctg_DriverUnload)
#endif


NTSTATUS DriverEntry(
                        IN PDRIVER_OBJECT DriverObject, 
                        IN PUNICODE_STRING RegistryPath
                    )
/*++

Routine Description:

    Installable driver initialization entry point.
    This entry point is called directly by the I/O system.

Arguments:

    DriverObject - pointer to the driver object

    RegistryPath - pointer to a unicode string representing the path,
                   to driver-specific key in the registry.

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise

--*/
{
    ULONG i;

    PAGED_CODE();

    UNREFERENCED_PARAMETER(RegistryPath);

    DBGOUT(("*** DriverEntry: build %s %s ***", __DATE__, __TIME__)); 

    /*
     *  Route all IRPs on device objects created by this driver
     *  to our IRP dispatch routine.
     */

    for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++){
        DriverObject->MajorFunction[i] = Ctg_Dispatch; 
    }

    DriverObject->DriverExtension->AddDevice = Ctg_AddDevice;
    DriverObject->DriverUnload = Ctg_DriverUnload;
//	DriverObject->DriverStartIo = Ctg_StartIo;
	DriverObject->DriverStartIo = NULL;

    return STATUS_SUCCESS;
}


NTSTATUS Ctg_AddDevice(
                        IN PDRIVER_OBJECT driverObj, 
                        IN PDEVICE_OBJECT physicalDevObj
                     )
/*++

Routine Description:

    The PlugPlay subsystem is handing us a brand new 
    PDO (Physical Device Object), for which we
    (by means of INF registration) have been asked to filter.

    We need to determine if we should attach or not.
    Create a filter device object to attach to the stack
    Initialize that device object
    Return status success.

    Remember: we can NOT actually send ANY non pnp IRPS to the given driver
    stack, UNTIL we have received an IRP_MN_START_DEVICE.

Arguments:

    driverObj - pointer to a device object.

    physicalDevObj -    pointer to a physical device object pointer 
                        created by the  underlying bus driver.

Return Value:

    NT status code.

--*/

{
    NTSTATUS status;
    PDEVICE_OBJECT filterDevObj = NULL;
    
    PAGED_CODE();

    DBGOUT(("Ctg_AddDevice: drvObj=%ph, pdo=%ph", driverObj, physicalDevObj)); 

    status = IoCreateDevice(    driverObj, 
                                sizeof(DEVICE_EXTENSION),
                                NULL,           // name for this device
                                FILE_DEVICE_UNKNOWN, //Oney says F_D_U not FILE_DEVICE_SERIAL_PORT, 
                                0,// Oney says 0 not FILE_AUTOGENERATED_DEVICE_NAME|FILE_DEVICE_SECURE_OPEN,                // device characteristics
                                FALSE,          // not exclusive
                                &filterDevObj); // our device object

    if (NT_SUCCESS(status)){
        PDEVICE_EXTENSION devExt;

        ASSERT(filterDevObj);

        /*
         *  Initialize device extension for new device object
         */
        devExt = (PDEVICE_EXTENSION)filterDevObj->DeviceExtension;
        RtlZeroMemory(devExt, sizeof(DEVICE_EXTENSION));
        devExt->signature = DEVICE_EXTENSION_SIGNATURE;
        devExt->state = STATE_INITIALIZED;
        devExt->filterDevObj = filterDevObj;
        devExt->physicalDevObj = physicalDevObj;
        
        devExt->pendingActionCount = 0;
        KeInitializeEvent(&devExt->removeEvent, NotificationEvent, FALSE);
#ifdef HANDLE_DEVICE_USAGE
        KeInitializeEvent(&devExt->deviceUsageNotificationEvent, SynchronizationEvent, TRUE);
#endif // HANDLE_DEVICE_USAGE


		// init DPC here (we'll use it not for ISR but for delayed event processing)
		KeInitializeTimer(&devExt->CtgTimer);

		IoInitializeDpcRequest(filterDevObj, Ctg_Dpc); 

		KeInitializeTimer(&devExt->ClrRtsTimer);

		KeInitializeDpc(&devExt->ClrRtsDpc, Ctg_ClrRtsDpc, devExt); 
		/*
		
		// section copied from Oney "wdm"

		status = IoRegisterDeviceInterface(filterDevObj, &GUID_CEC_CTG, NULL, &devExt->ifname);
		if (!NT_SUCCESS(status))
		{						// unable to register interface
			KdPrint(("CEC_CTG driver - "
				"IoRegisterDeviceInterface failed - %8.8lX\n", status));
			IoDeleteDevice(filterDevObj);
			return status;
		}						// unable to register interface
		IoSetDeviceInterfaceState(&devExt->ifname, TRUE);
		KdPrint(("CEC_CTG driver - "
			"created interface named %ws\n", devExt->ifname.Buffer));

		*/

        /*
         *  Attach the new device object to the top of the device stack.
         */
        devExt->topDevObj = IoAttachDeviceToDeviceStack(filterDevObj, physicalDevObj);

        ASSERT(devExt->topDevObj);
        DBGOUT(("created filterDevObj %ph attached to %ph.", filterDevObj, devExt->topDevObj));


        //
        // As a filter driver, we do not want to change the power or I/O
        // behavior of the driver stack in any way.  Recall that a filter
        // driver should "appear" the same (almost) as the underlying device.
        // Therefore we must copy some bits from the device object _directly_
        // below us in the device stack (notice: DON'T copy from the PDO!)
        //

        /* Various I/O-related flags which should be maintained */
        /* (copy from lower device object) */
        filterDevObj->Flags |=
            (devExt->topDevObj->Flags & (DO_BUFFERED_IO | DO_DIRECT_IO));

        /* Various Power-related flags which should be maintained */
        /* (copy from lower device object) */
        filterDevObj->Flags |= (devExt->topDevObj->Flags &
            (DO_POWER_INRUSH | DO_POWER_PAGABLE /*| DO_POWER_NOOP*/)); 


		// from Oney's book
		filterDevObj->DeviceType = devExt->topDevObj->DeviceType;
		filterDevObj->Characteristics = devExt->topDevObj->Characteristics;

/*
		// init DPC 
		IoInitializeDpcRequest(filterDevObj, Ctg_DpcForIsr);
*/

#ifdef HANDLE_DEVICE_USAGE
        //
        // To determine whether some of our routines should initially be
        // pageable, we must consider the DO_POWER_xxxx flags of the
        // device object directly below us in the device stack.
        //
        // * We make ourselves pageable if:
        //     - that devobj has its PAGABLE bit set (so we know our power
        //       routines won't be called at DISPATCH_LEVEL)
        // -OR-
        //     - that devobj has its NOOP bit set (so we know we won't be
        //       participating in power-management at all).  NOTE, currently
        //       DO_POWER_NOOP is not implemented.
        //
        // * Otherwise, we make ourselves non-pageable because either:
        //     - that devobj has its INRUSH bit set (so we also have to be
        //       INRUSH, and code that handles INRUSH irps can't be pageable)
        // -OR-
        //     - that devobj does NOT have its PAGABLE bit set (and NOOP isn't
        //       set, so some of our code might be called at DISPATCH_LEVEL)
        //
        if ((devExt->topDevObj->Flags & DO_POWER_PAGABLE)
             /*|| (devExt->topDevObj->Flags & DO_POWER_NOOP)*/)
        {
            // We're initially pageable.
            //
            // Don't need to do anything else here, for now.
        }
        else
        {
            // We're initially non-pageable.
            //
            // We need to lock-down the code for all routines
            // that could be called at IRQL >= DISPATCH_LEVEL.
            DBGOUT(( "LOCKing some driver code (non-pageable) (b/c init conditions)" ));
            devExt->initUnlockHandle = MmLockPagableCodeSection( Ctg_Power );  // some func that's inside the code section that we want to lock
            ASSERT( NULL != devExt->initUnlockHandle );
        }

        /*
         *  Remember our initial flag settings.
         *  (Need remember initial settings to correctly handle
         *  setting of PAGABLE bit later.)
         */
        devExt->initialFlags = filterDevObj->Flags & ~DO_DEVICE_INITIALIZING;
#endif // HANDLE_DEVICE_USAGE

        /*
         *  Clear the initializing bit from the new device object's flags.
         *  NOTE: must not do this until *after* setting DO_POWER_xxxx flags
         */
        filterDevObj->Flags &= ~DO_DEVICE_INITIALIZING;

        /*
         *  This is a do-nothing call to a sample function which
         *  demonstrates how to read the device's registry area.
         *  Note that you cannot make this call on devExt->filterDevObj
         *  because a filter device object does not have a devNode.
         *  We pass devExt->physicalDevObj, which is the device object
         *  for which this driver is a filter driver.
         */
        RegistryAccessSample(devExt, devExt->physicalDevObj);
    } 

    ASSERT(NT_SUCCESS(status));
    return status;
}


VOID Ctg_DriverUnload(IN PDRIVER_OBJECT DriverObject)
/*++

Routine Description:

    Free all the allocated resources, etc.

    Note:  Although the DriverUnload function often does nothing,
           the driver must set a DriverUnload function in 
           DriverEntry; otherwise, the kernel will never unload
           the driver.

Arguments:

    DriverObject - pointer to a driver object.

Return Value:

    VOID.

--*/
{
    PAGED_CODE();

    DBGOUT(("Ctg_DriverUnload")); 

/*    PDEVICE_EXTENSION devExt;

    ASSERT(filterDevObj);

    devExt = (PDEVICE_EXTENSION)filterDevObj->DeviceExtension;

	RtlFreeUnicodeString (???);

	*/
}


NTSTATUS Ctg_Dispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
/*++

Routine Description:

    Common entrypoint for all Io Request Packets

Arguments:

    DeviceObject - pointer to a device object.
    Irp - Io Request Packet

Return Value:

    NT status code.

--*/

{
    PDEVICE_EXTENSION devExt;
    PIO_STACK_LOCATION irpSp;
    BOOLEAN passIrpDown = TRUE;
    UCHAR majorFunc, minorFunc;
    NTSTATUS status = STATUS_SUCCESS;

    devExt = DeviceObject->DeviceExtension;
    ASSERT(devExt->signature == DEVICE_EXTENSION_SIGNATURE);

    irpSp = IoGetCurrentIrpStackLocation(Irp);

    /*
     *  Get major/minor function codes in private variables
     *  so we can access them after the IRP is completed.
     */
    majorFunc = irpSp->MajorFunction;
    minorFunc = irpSp->MinorFunction;

//    DBGOUT(("Ctg_Dispatch: majorFunc=%d (%s), minorFunc=%d\n", 
//            (ULONG)majorFunc, IrpMjFuncDesc[majorFunc], (ULONG)minorFunc)); 

    /*
     *  For all IRPs except REMOVE, we increment the PendingActionCount
     *  across the dispatch routine in order to prevent a race condition with
     *  the REMOVE_DEVICE IRP (without this increment, if REMOVE_DEVICE
     *  preempted another IRP, device object and extension might get
     *  freed while the second thread was still using it).
     */
    if (!((majorFunc == IRP_MJ_PNP) && (minorFunc == IRP_MN_REMOVE_DEVICE))){
        IncrementPendingActionCount(devExt);
    }

    if ((majorFunc != IRP_MJ_PNP) &&
        (majorFunc != IRP_MJ_CLOSE) &&
        ((devExt->state == STATE_REMOVING) ||
         (devExt->state == STATE_REMOVED))){

        /*
         *  While the device is being removed, 
         *  we only pass down the PNP and CLOSE IRPs.
         *  We fail all other IRPs.
         */
        status = Irp->IoStatus.Status = STATUS_DELETE_PENDING;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
        passIrpDown = FALSE;
    }
    else {
        switch (majorFunc){

            case IRP_MJ_PNP:
                status = Ctg_PnP(devExt, Irp);
                passIrpDown = FALSE;
                break;

            case IRP_MJ_POWER:
                status = Ctg_Power(devExt, Irp);
                passIrpDown = FALSE;
                break;

				// process our requests
            case IRP_MJ_DEVICE_CONTROL:
				{
				ULONG info = 0;
				// Second parameter in DeviceIoControl() call
				ULONG code = irpSp->Parameters.DeviceIoControl.IoControlCode;

				//Length of the output buffersixth argument to DeviceIoControl()
				ULONG outlength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
				
				//Length of the input bufferfourth argument to DeviceIoControl()
				ULONG inlength = irpSp->Parameters.DeviceIoControl.InputBufferLength;

				// in METHOD_BUFFERED operations all the data are located in SystemBuffer.
				PVOID pSystemBuffer = Irp->AssociatedIrp.SystemBuffer;

//				DBGOUT(("Ctg_Dispatch: IRP_MJ_DEVICE_CONTROL, command = %d (%s), Irp=%x\n", code, 
//					(code==IOCTL_CTG_ON)?"IOCTL_CTG_ON":((code==IOCTL_CTG_OFF)?"IOCTL_CTG_OFF": \
//				((code==IOCTL_CTG_VERSION)?"IOCTL_CTG_VERSION":"Unknown!")), Irp)); 

				switch (code)
				{						// process control operation
				case IOCTL_CTG_ON:
					{
					NTSTATUS ntStatus;
					
					DBGOUT(("Starting CEC_CTG, address = %d, DelayAfterFirstByte = %d, DelayAfterRTSSet = %d\n", 
						devExt->RVUAddress, devExt->driver_parameters.DelayAfterFirstByte, 
						devExt->driver_parameters.DelayAfterRtsSet));

					devExt->bCtgDriverIsOn = TRUE; // filter is initially silent
//					DBGOUT(("Ctg_Dispatch: bCtgDriverIsOn = %d\n", devExt->bCtgDriverIsOn)); 

					KeSetTimerEx(&devExt->CtgTimer, 
						RtlConvertLongToLargeInteger(-(LONG)(1000000)), // 100 ms
						100, // 100 ms period
						&devExt->filterDevObj->Dpc);

// now start waiting for received char
					// Setup call for device control to wait for a serial event
					
					// if some process is pending now we will not interfere it
					if(devExt->pPendingIrp != NULL)
					{
						status = Irp->IoStatus.Status = STATUS_PENDING;
						Irp->IoStatus.Information = info;
						IoCompleteRequest(Irp, IO_NO_INCREMENT);

						passIrpDown = FALSE;
						break;
					}

					Ctg_PurgeRx(devExt); // clear input before waiting for events

					if(!NT_SUCCESS(Ctg_SetWaitMask(devExt, SERIAL_EV_RXCHAR)))
					{
						status = STATUS_INSUFFICIENT_RESOURCES;
						passIrpDown = FALSE;
						//return STATUS_INVALID_BUFFER_SIZE; -- is invalid here since won't release
						// pending action counter at the end
						break;
					}

					ntStatus = StartWaitOnMask(devExt);
					// completing original IRP
//					DBGOUT(("Ctg_Dispatch IOCTL_CTG_ON: completing IRP %x.\n", Irp));

					status = Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = info;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;
					}
					break;
				case IOCTL_CTG_OFF:
					{
					NTSTATUS ntStatus;
					devExt->bCtgDriverIsOn = FALSE; // filter is initially silent

					Ctg_ClrRTS(devExt);
					devExt->bRTSIsOn = FALSE;

					// cancel waiting on mask
//					devExt->pIOCtlBuffer[0]=0;
//					ntStatus = SendIoctlSerialImmediate(devExt, IOCTL_SERIAL_SET_WAIT_MASK, sizeof(ULONG));
//					if(!NT_SUCCESS(ntStatus))
//					{
//						DBGOUT(("Error %x (%d) calling SendIoctlSerialImmediate!\n", ntStatus, ntStatus));
//					}
					
//					DBGOUT(("Ctg_Dispatch: bCtgDriverIsOn = %d\n", devExt->bCtgDriverIsOn)); 

					status = Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = info;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;

					// cancel pending IOCTL if any
					if(devExt->pPendingIrp != NULL)
					{
						// DBGOUT(("Cancelling IRP %x...\n", devExt->pPendingIrp));
						/*
						devExt->pPendingIrp->IoStatus.Status = STATUS_CANCELLED;
						IoCompleteRequest(devExt->pPendingIrp, IO_NO_INCREMENT);
						*/
						
						// driver will execute this IRP and cancel it himself?
						// IoCancelIrp(devExt->pPendingIrp);
						
						//devExt->pPendingIrp = NULL;
					}
					}
					break;
				case IOCTL_CTG_VERSION:
					{						
					PULONG pversion;
/*					KeSetTimer(&devExt->ClrRtsTimer, 
						RtlConvertLongToLargeInteger(-(LONG)(20000000)), // 1000 us  10000
						&devExt->ClrRtsDpc); // debug*/
					if (outlength < sizeof(ULONG))
					{						// invalid buffer size
						DBGOUT(("Ctg_Dispatch: outlength < sizeof(ULONG), outlength = %d\n", outlength)); 
						status = Irp->IoStatus.Status = STATUS_INVALID_BUFFER_SIZE;
						Irp->IoStatus.Information = 0;//info;
						IoCompleteRequest(Irp, IO_NO_INCREMENT);

						passIrpDown = FALSE;
						//return STATUS_INVALID_BUFFER_SIZE; -- is invalid here since won't release
						// pending action counter at the end
						break;
					}						// invalid buffer size

					DBGOUT(("Ctg_Dispatch: IOCTL_CTG_VERSION, build:%s", __TIME__)); 

					pversion = (PULONG) pSystemBuffer;
					info = sizeof(ULONG); // how many bytes are to be copied after IRP exits
					*pversion = DRIVER_VERSION;

//					DBGOUT(("Ctg_Dispatch: pversion = (PULONG)\n")); 

					status = Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = info;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;
					}
					break;
				case IOCTL_GET_DIAG_INFO:
					{
					PDIAG_INFO pDiagInfo;

					if (outlength < sizeof(DIAG_INFO) )
					{						// invalid buffer size
						DBGOUT(("Ctg_Dispatch: outlength < sizeof(DIAG_INFO)\n", outlength)); 
						status = Irp->IoStatus.Status = STATUS_INVALID_BUFFER_SIZE;
						Irp->IoStatus.Information = 0;//info;
						IoCompleteRequest(Irp, IO_NO_INCREMENT);
						passIrpDown = FALSE;
						//return STATUS_INVALID_BUFFER_SIZE; -- is invalid here since won't release
						// pending action counter at the end
						break;
					}						// invalid buffer size

					// set some parameters here
					devExt->diag_info.nPacketsInTxQueue = devExt->NextTxPktWrite - devExt->NextTxPktRead; 
					if(devExt->diag_info.nPacketsInTxQueue > TX_PKT_ARRAY_SIZE) // ULONG < 0
						devExt->diag_info.nPacketsInTxQueue += TX_PKT_ARRAY_SIZE;

					devExt->diag_info.nPacketsInRxQueue = devExt->NextRxPktWrite - devExt->NextRxPktRead; 
					if(devExt->diag_info.nPacketsInRxQueue > RX_PKT_ARRAY_SIZE)// ULONG < 0
						devExt->diag_info.nPacketsInRxQueue += RX_PKT_ARRAY_SIZE;

					pDiagInfo = (PDIAG_INFO) pSystemBuffer;
					info = sizeof(DIAG_INFO);
					RtlMoveMemory(pDiagInfo, &devExt->diag_info, sizeof(DIAG_INFO));
	
//					DBGOUT(("Ctg_Dispatch: IOCTL_GET_DIAG_INFO:info=%d", info)); 

					status = Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = info;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;
					}
					break;
				case IOCTL_SET_DRIVER_PARAMETERS:
					{						
					PDRIVER_PARAMETERS pDrvParm;

					if (inlength != sizeof(DRIVER_PARAMETERS) )
					{						// invalid buffer size
						DBGOUT(("Ctg_Dispatch: inlength != sizeof(DIAG_INFO)\n", inlength)); 
						status = Irp->IoStatus.Status = STATUS_INVALID_BUFFER_SIZE;
						Irp->IoStatus.Information = 0;//info;
						IoCompleteRequest(Irp, IO_NO_INCREMENT);
						passIrpDown = FALSE;
						//return STATUS_INVALID_BUFFER_SIZE; -- is invalid here since won't release
						// pending action counter at the end
						break;
					}						// invalid buffer size

					pDrvParm = (PDRIVER_PARAMETERS) pSystemBuffer;
					info = 0;
					// check parameter validity
					if(pDrvParm->DelayAfterFirstByte > 40 && pDrvParm->DelayAfterFirstByte < 5000)
						devExt->driver_parameters.DelayAfterFirstByte = pDrvParm->DelayAfterFirstByte;
					if(pDrvParm->DelayAfterRtsSet > 40 && pDrvParm->DelayAfterRtsSet < 5000)
						devExt->driver_parameters.DelayAfterRtsSet = pDrvParm->DelayAfterRtsSet;
					// set test mode on or off
					devExt->driver_parameters.bTestMode = pDrvParm->bTestMode;
					// set RVU address from outside if valid
					if(pDrvParm->RVUAddress >= 211 && pDrvParm->RVUAddress <= 213)
						devExt->RVUAddress = pDrvParm->RVUAddress;
					// change ACK packet correspondingly
					Ctg_InitACKPacket(devExt, pDrvParm->bTestMode);

					status = Irp->IoStatus.Status = STATUS_SUCCESS;
					Irp->IoStatus.Information = info;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;
					}
					break;
				default:
					break;
				}						// process control operation


				}
				break;

            case IRP_MJ_CREATE:
				Ctg_InitStorage(devExt);
				// passIrpDown = TRUE;
				break;
            case IRP_MJ_CLOSE:
				devExt->bCtgDriverIsOn = FALSE;

				// it could be a good idea to call IOCTL_CTG_OFF...
				Ctg_ClrRTS(devExt);
				devExt->bRTSIsOn = FALSE;
				// passIrpDown = TRUE;
				break;
            case IRP_MJ_READ:
				if(devExt->bCtgDriverIsOn == TRUE)
				{
					ULONG nBytesRead = 0;
					PPACKET pPkt = NULL;

					// read data to buffer
//					DBGOUT(("IRP_MJ_READ: want to read %d bytes!", irpSp->Parameters.Read.Length));
					
					// look if meaningful packet for reading exists
					if(devExt->NextRxPktRead != devExt->NextRxPktWrite)
					{
						pPkt = devExt->arRxPkt + devExt->NextRxPktRead;
						nBytesRead = pPkt->dataLength;
						if(devExt->NextRxPktRead+1 >= RX_PKT_ARRAY_SIZE)
							devExt->NextRxPktRead = 0;
						else
							devExt->NextRxPktRead++;
					}
					else 
						nBytesRead = 0;

					// check if target buffer is sufficient 
					if(irpSp->Parameters.Read.Length < nBytesRead)
					{
						status = Irp->IoStatus.Status = STATUS_INVALID_BUFFER_SIZE;
					}
					else
					{
						status = Irp->IoStatus.Status = STATUS_SUCCESS;
						if(nBytesRead && pPkt != NULL)
							RtlMoveMemory(Irp->AssociatedIrp.SystemBuffer, pPkt, pPkt->dataLength);
					}
					Irp->IoStatus.Information = nBytesRead;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;
				}
				break;

            case IRP_MJ_WRITE:
				if(devExt->bCtgDriverIsOn == TRUE)
				{
					ULONG nBytesWrite = irpSp->Parameters.Write.Length;
					BOOLEAN bBufferFull = FALSE;
					BOOLEAN bBadPacket = FALSE;
					PUCHAR pByte = Irp->AssociatedIrp.SystemBuffer;

//					DBGOUT(("IRP_MJ_WRITE: want to write %d bytes!", nBytesWrite));

					// check if send buffer is full
					if(devExt->NextTxPktWrite == (TX_PKT_ARRAY_SIZE-1) && devExt->NextTxPktRead == 0)
						bBufferFull = TRUE;
					else if(devExt->NextTxPktWrite == devExt->NextTxPktRead-1)
						bBufferFull = TRUE;
					if(nBytesWrite > 255) // data too long
						bBadPacket = TRUE;

					// HERE WAS A BUG!!! You cannot access any buffer if it is invalid!!!
					if(nBytesWrite == 0) // no data, buffer is invalid
						bBadPacket = TRUE; // avoid touching this buffer!

					// check if data is a valid packet
					// HERE WAS A BUG!!! You cannot access any buffer if it is invalid!!!
					if(bBadPacket == FALSE)
					{
						if(*pByte++ != 1)
						{
							bBadPacket = TRUE;
							DBGOUT(("IRP_MJ_WRITE: Byte 1 is not 1, packet is bad!"));
						}
					}
					if(bBadPacket == FALSE)
					{
						if(*pByte++ != devExt->RVUAddress)
						{
							bBadPacket = TRUE;
							DBGOUT(("IRP_MJ_WRITE:  Byte 2 is not devExt->RVUAddress, packet is bad!"));
						}
					}
					if(bBadPacket == FALSE)
					{
						if(*pByte++ != nBytesWrite-4) // length
						{
							bBadPacket = TRUE;
							DBGOUT(("IRP_MJ_WRITE: data length is not nBytesWrite+4, packet is bad!"));
						}
					}

					// read data to buffer
					if(bBufferFull == TRUE || bBadPacket == TRUE)
					{ // more than 1 packet is copied
						status = Irp->IoStatus.Status = STATUS_INVALID_BUFFER_SIZE;
						nBytesWrite = 0;
					}
					else
					{
//						DBGOUT(("IRP_MJ_WRITE: Hmm, packet is good..."));
						status = Irp->IoStatus.Status = STATUS_SUCCESS;
						RtlMoveMemory(devExt->arTxPkt[devExt->NextTxPktWrite].data, Irp->AssociatedIrp.SystemBuffer, nBytesWrite);
						devExt->arTxPkt[devExt->NextTxPktWrite].bPacketValid = TRUE;
						devExt->arTxPkt[devExt->NextTxPktWrite].dataLength = nBytesWrite;
						devExt->arTxPkt[devExt->NextTxPktWrite].Length = nBytesWrite - 4;
						devExt->arTxPkt[devExt->NextTxPktWrite].Position = 0;
						devExt->arTxPkt[devExt->NextTxPktWrite].Checksum = 0;
						// increase next write packet address

						if(devExt->NextTxPktWrite+1 >=TX_PKT_ARRAY_SIZE)
							devExt->NextTxPktWrite = 0;
						else
							devExt->NextTxPktWrite++;

					}

					Irp->IoStatus.Information = nBytesWrite;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);

					passIrpDown = FALSE;
//					DBGOUT(("After IRP_MJ_WRITE"));
				}
				break;

            case IRP_MJ_SYSTEM_CONTROL:
            case IRP_MJ_INTERNAL_DEVICE_CONTROL:
            default:
                /*
                 *  For unsupported IRPs, we simply send the IRP
                 *  down the driver stack.
                 */
                break;
        }
    }

    if (passIrpDown){
        IoCopyCurrentIrpStackLocationToNext(Irp);
        status = IoCallDriver(devExt->topDevObj, Irp);
    }

    /*
     *  Balance the increment to PendingActionCount above.
     */
    if (!((majorFunc == IRP_MJ_PNP) && (minorFunc == IRP_MN_REMOVE_DEVICE))){
        DecrementPendingActionCount(devExt);
    }

    return status;
}

NTSTATUS OnStartWaitMaskIrpComplete(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context)
{
	PULONG pMask; // mask of event that completed this IRP
	ULONG uMask;

	NTSTATUS ntStatus;
	BOOLEAN bNeedWaitEvent = TRUE;
	BOOLEAN bNeedWritePort = FALSE;

	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) Context;
//	DBGOUT(("OnStartWaitMaskIrpComplete: Irp (%x, status=%x, info=%d).", Irp, Irp->IoStatus.Status, Irp->IoStatus.Information));

	pMask = Irp->AssociatedIrp.SystemBuffer;
	uMask = *pMask;

	// if driver is on and IRP is not cancelled
	// Irp->Cancel is TRUE when IRP is cancelled from outside.
	if(Irp->Cancel == FALSE && pDevExt->bCtgDriverIsOn == TRUE)
	{
		Ctg_SetWaitMask(pDevExt, SERIAL_EV_RXCHAR); // wait for next char if not cancelled

		if(uMask == SERIAL_EV_RXCHAR) // next char received from serial port
		{
			ULONG nRead = 0;
			BOOLEAN bPollPacket = FALSE;
			BOOLEAN bDiagInfo_NotPacket57 = FALSE;

	//		DBGOUT(("uMask == SERIAL_EV_RXCHAR"));
			
			// read the char(s)
			ntStatus = ReadPort(pDevExt);

			Ctg_SetWaitMask(pDevExt, SERIAL_EV_RXCHAR);
				
			// process chars
			while(pDevExt->RxBufferRawReadIndex != pDevExt->RxBufferRawWriteIndex)
			{
				PPACKET pPkt = pDevExt->pRxPktCurr;

				UCHAR NextByte = pDevExt->pRxBufferRaw[pDevExt->RxBufferRawReadIndex];
	//			DBGOUT(("** NextByte = %d (%x, %c)", NextByte, NextByte, NextByte));
				if(pPkt != NULL)
				{
					if(pPkt->bPacketStarted == FALSE)
					{
						if(NextByte == pDevExt->RVUAddress) // added in v.1.0.0.4 for other RVUs
						{
	//						DBGOUT(("Packet started!"));
							pPkt->bPacketStarted = TRUE;
							pPkt->Position = 0;
							pPkt->Length = 0;
							pPkt->Checksum = 0;
						}
					}
					// check bytes 1 and 2
					if(pPkt->bPacketStarted == TRUE)
					{
						if(pPkt->Position == 1) // from
						{
							if(NextByte != 1) // bad second byte
							{
//								DBGOUT(("Second byte is not 1, packet is bad!"));
								pPkt->bPacketStarted = FALSE;
								pPkt->Position = 0;
								pPkt->Length = 0;
								pPkt->Checksum = 0;
							}
						}
						if(pPkt->Position == 2) // length
						{
							pPkt->Length = NextByte;
						}
					}

					if(pPkt->bPacketStarted == TRUE)
					{
						if(pPkt->Position<254)
						{
							pPkt->data[pPkt->Position]=NextByte;
							if(pPkt->Position++ < pPkt->Length+3)
								pPkt->Checksum += NextByte;
							else // time to check if checksums match
							{
								if( (pPkt->Checksum & 0xff) == NextByte)
								{
									pPkt->bPacketValid = TRUE;
									// pDevExt->RVUAddress = pPkt->data[0]; // don't remember RVU address on this port, set it externally
									// pDevExt->pktACKPacket.data[1] = pDevExt->RVUAddress; // save this address in test packet
									bPollPacket = TRUE; // we received a valid packet
									if(pPkt->data[3] != 57) // did we receive a poll packet or a meaningful one?
										bDiagInfo_NotPacket57 = TRUE;
									pPkt->dataLength = pPkt->Length+4;
								}
								else // we received a bad packet
								{
									pPkt->bPacketStarted = FALSE;
									pPkt->Position = 0;
									pPkt->Length = 0;
									pPkt->Checksum = 0;
								}
							}
						}//if(pPkt->Position<254)
					} //if(pPkt->bPacketStarted == TRUE)
				} //if(pPkt != NULL)

				// the actual value should NEVER be set to a bad value!!!
				if(pDevExt->RxBufferRawReadIndex+1 >= RX_BUFFER_RAW_SIZE)
					pDevExt->RxBufferRawReadIndex = 0;
				else
					pDevExt->RxBufferRawReadIndex++;
			}

			if(bPollPacket == TRUE)
			{
				BOOLEAN bParity = TRUE;

	//			DBGOUT(("Poll packet found!"));
				// if have packets to send - send the next!
				if(pDevExt->NextTxPktRead != pDevExt->NextTxPktWrite)
				{
					ULONG count=0;
					while(pDevExt->NextTxPktRead != pDevExt->NextTxPktWrite)
					{
						BOOLEAN bNeedBreak = FALSE;

	//					DBGOUT(("pDevExt->NextTxPktRead=%d, pDevExt->NextTxPktWrite=%d", 
	//						pDevExt->NextTxPktRead, pDevExt->NextTxPktWrite));

						ASSERT(pDevExt->pTxPktCurr == NULL);

						// validate parameters
						if(pDevExt->arTxPkt[pDevExt->NextTxPktRead].bPacketValid)
						{
							pDevExt->pTxPktCurr = &pDevExt->arTxPkt[pDevExt->NextTxPktRead];
							//DBGOUT(("pDevExt->pTxPktCurr=pDevExt->arTxPkt+%d", pDevExt->NextTxPktRead));

							bNeedBreak = TRUE;
						}

						if(pDevExt->NextTxPktRead+1 >= TX_PKT_ARRAY_SIZE)
							pDevExt->NextTxPktRead = 0;
						else
							pDevExt->NextTxPktRead++;
						
						if(bNeedBreak)
							break;
						if(++count > TX_PKT_ARRAY_SIZE)
							break;
					}
				}
				else // nothing to send - send ACK
				{
					// reset position in ACK packet to 0
					pDevExt->pktACKPacket.Position = 0;
//					pDevExt->pktACKPacket.data[1] = pDevExt->RVUAddress; there can be 3 RVU addresses from in v.1.0.0.4
					pDevExt->pTxPktCurr = &pDevExt->pktACKPacket; // ACK packet address
				}
				// now we have valid pDevExt->pTxPktCurr.
				// initialize transmitting (send first byte; next bytes will be sent 
				// one by one in OnWriteComplete
//				ntStatus = WritePort(pDevExt);
				bNeedWritePort = TRUE;
				
				// add diagnostics
				pDevExt->diag_info.nTotalPacketsTransmitted++;
				if(pDevExt->pTxPktCurr->data[3] != 12) // meaningful packet sent
					pDevExt->diag_info.nMeaningfulPacketsTransmitted++;
				
				// set next placeholder as a target for next packet writing
				// if this packet is not a simple poll. Otherwise write next packet to the same location
				if(pDevExt->pRxPktCurr->data[3] != 57)
				{
					if(pDevExt->NextRxPktWrite+1 >= RX_PKT_ARRAY_SIZE)
						pDevExt->NextRxPktWrite = 0;
					else
						pDevExt->NextRxPktWrite++;

					pDevExt->pRxPktCurr = pDevExt->arRxPkt+pDevExt->NextRxPktWrite;
				}
				Ctg_ClearPacket(pDevExt->pRxPktCurr);

//				DBGOUT(("Next packet received!"));
				
				// add diagnostics
				pDevExt->diag_info.nTotalPacketsReceived++;
				if(bDiagInfo_NotPacket57)
					pDevExt->diag_info.nMeaningfulPacketsReceived++;
				pDevExt->diag_info.TimeoutSinceLastPacketReceived = 0; // clear timeout
			}
			
			bNeedWaitEvent = TRUE;
		}
		else
		{
	//		DBGOUT(("OnStartWaitMaskIrpComplete: unknown uMask=%d, Cancel=%d!Did not set wait event!", uMask, Irp->Cancel));
		}
	} //if(Irp->Cancel == FALSE)

	// clear operation if driver is off
	if(pDevExt->bCtgDriverIsOn == FALSE)
		bNeedWaitEvent = bNeedWritePort = FALSE;

	// Now, only AFTER we read data from IRP, we can release it!

	if(Irp == pDevExt->pPendingIrp)
	{
		// we have no pending IRPs more
		pDevExt->pPendingIrp = NULL;
		
//		DBGOUT(("Irp == pDevExt->pPendingIrp, freeing it!\n"));

		IoFreeIrp(Irp);
	}

	if(bNeedWritePort == TRUE)
		ntStatus = WritePort(pDevExt);

	if(bNeedWaitEvent == TRUE && Irp->Cancel == FALSE) // continuing work if IRP is not cancelled
	{
		ntStatus = StartWaitOnMask(pDevExt);
	}

	return STATUS_MORE_PROCESSING_REQUIRED;
}


// Frees IRP; used with IoAllocateIrp
NTSTATUS OnEmptyIrpCompleteFree(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context)
{
	PULONG pMask;
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) Context;
//	DBGOUT(("OnEmptyIrpCompleteFree (%x) completed (status=%x, info=%d).", Irp, Irp->IoStatus.Status, Irp->IoStatus.Information));

	pMask = Irp->AssociatedIrp.SystemBuffer;

	IoFreeIrp(Irp);

	return STATUS_MORE_PROCESSING_REQUIRED;
}

// Does not free IRP; used with KeWaitSync
NTSTATUS OnEmptyIrpComplete(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) Context;
//	DBGOUT(("OnEmptyIrpComplete (%x) completed (status=%x, info=%d).", Irp, Irp->IoStatus.Status, Irp->IoStatus.Information));

//	IoFreeIrp(Irp);

	return STATUS_MORE_PROCESSING_REQUIRED;
}


NTSTATUS OnWriteIrpComplete(PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context)
{
//	PUCHAR pByte;
	ULONG i;

	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) Context;

	if(pDevExt == NULL)
	{
//		DBGOUT(("OnWriteIrpComplete: pDevExt == NULL!"));
	}
//	DBGOUT(("OnWriteIrpComplete: (%x) completed (status=%x, info=%d).", Irp, Irp->IoStatus.Status, Irp->IoStatus.Information));

	IoFreeIrp(Irp);

	WritePort(pDevExt);

	return STATUS_MORE_PROCESSING_REQUIRED;
}

// Function is used for sending a comamns to serial port that returns immediately
// (like IOCTL_SERIAL_CLR_RTS or IOCTL_SERIAL_SET_RTS or IOCTL_SERIAL_SET_WAIT_MASK)
NTSTATUS SendIoctlSerialImmediate(PDEVICE_EXTENSION pDevExt, ULONG Request, ULONG nBytesToSend)
{
	PIRP Irp;
	NTSTATUS ntStatus = STATUS_SUCCESS;
	KEVENT evIrpComplete;
	
	Irp = IoAllocateIrp(pDevExt->filterDevObj->StackSize, FALSE);

	KeInitializeEvent( &evIrpComplete, SynchronizationEvent, FALSE );
	
	if(Irp == NULL)
	{
//		DBGOUT(("SendIoctlSerialImmediate: Cannot allocate IRP! \n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		PIO_STACK_LOCATION stack = IoGetNextIrpStackLocation(Irp);

		stack->MajorFunction = IRP_MJ_DEVICE_CONTROL;
		stack->Parameters.DeviceIoControl.IoControlCode = Request;
		stack->Parameters.DeviceIoControl.InputBufferLength = nBytesToSend;
		stack->Parameters.DeviceIoControl.OutputBufferLength = nBytesToSend;
		Irp->AssociatedIrp.SystemBuffer = pDevExt->pIOCtlBuffer;

		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)OnEmptyIrpCompleteFree, (PVOID) pDevExt, TRUE, TRUE, TRUE);
		
		ntStatus = IoCallDriver(pDevExt->topDevObj, Irp);
		
//		DBGOUT(("SendIoctlSerialImmediate: IoCallDriver returns %d (%s)\n", ntStatus, 
//			(ntStatus==STATUS_PENDING)?"STATUS_PENDING":((ntStatus==STATUS_SUCCESS)?"STATUS_SUCCESS":"Unknown!"))); 
		
		if(ntStatus == STATUS_PENDING)
		{
			// Wait for IRP completion
//			DBGOUT(("SendIoctlSerialImmediate: Waiting KeWaitForSingleObject\n", ntStatus, ntStatus));
			KeWaitForSingleObject( &evIrpComplete, Executive, KernelMode, FALSE, NULL );
			// IoCompleteRequest is done inside
		}
		else if (!NT_SUCCESS(ntStatus))
		{
			DBGOUT(("SendIoctlSerialImmediate: Error using IoCallDriver: ntStatus=%x (%d)!\n", ntStatus, ntStatus));
		}
	}
	return ntStatus;
}


// Function is used for sending a IOCTL_SERIAL_WAIT_ON_MASK request that is pending
// OnStartWaitMaskIrpComplete is set.
//
// Important! You should call SendIoctlSerialImmediate with IOCTL_SERIAL_SET_WAIT_MASK before calling this function!
//

NTSTATUS StartWaitOnMask(PDEVICE_EXTENSION pDevExt)
{
	PIRP Irp;
	NTSTATUS ntStatus = STATUS_SUCCESS;
	
	Irp = IoAllocateIrp(pDevExt->filterDevObj->StackSize, FALSE);

	if(Irp == NULL)
	{
		DBGOUT(("StartWaitOnMask: Cannot allocate IRP! \n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		PIO_STACK_LOCATION stack = IoGetNextIrpStackLocation(Irp);

		stack->MajorFunction = IRP_MJ_DEVICE_CONTROL;
		stack->Parameters.DeviceIoControl.IoControlCode = IOCTL_SERIAL_WAIT_ON_MASK;
		stack->Parameters.DeviceIoControl.OutputBufferLength = sizeof(ULONG);
		Irp->AssociatedIrp.SystemBuffer = pDevExt->pIOCtlBuffer;

//		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)OnEmptyIrpCompleteFree, (PVOID) pDevExt, TRUE, TRUE, TRUE);
		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)OnStartWaitMaskIrpComplete, (PVOID) pDevExt, TRUE, TRUE, TRUE);
		
		ntStatus = IoCallDriver(pDevExt->topDevObj, Irp);
		
//		DBGOUT(("StartWaitOnMask: IoCallDriver returns %d (%s)\n", ntStatus, 
//			(ntStatus==STATUS_PENDING)?"STATUS_PENDING":((ntStatus==STATUS_SUCCESS)?"STATUS_SUCCESS":"Unknown!"))); 
		
		if(ntStatus == STATUS_PENDING)
		{
			// Wait for IRP completion
//			DBGOUT(("StartWaitOnMask: ntStatus == STATUS_PENDING\n"));
			pDevExt->pPendingIrp = Irp;
		}
		else if (!NT_SUCCESS(ntStatus))
		{
			DBGOUT(("StartWaitOnMask: Error using IoCallDriver: ntStatus=%x (%d)!\n", ntStatus, ntStatus));
		}
	}
	return ntStatus;
}


// Runs at DISPATCH_LEVEL!
VOID Ctg_Dpc(IN PKDPC Dpc, IN PDEVICE_OBJECT fdo, IN PIRP junk, IN PVOID Context)
{
	PDEVICE_EXTENSION pDevExt = fdo->DeviceExtension;

	pDevExt->diag_info.TimeoutSinceLastPacketReceived++;

	ASSERT(KeGetCurrentIrql() <= DISPATCH_LEVEL);

	if(pDevExt->bCtgDriverIsOn == FALSE)
		KeCancelTimer(&pDevExt->CtgTimer);
//	DBGOUT(("Ctg_Dpc is called."));
//	DBGOUT(("Ctg_Dpc is called, calling ReadPort."));
//	ReadPort(pDevExt);
}

// Runs at DISPATCH_LEVEL!
// Intended for clearing RTS after delay (on timer)
VOID Ctg_ClrRtsDpc(IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
	PDEVICE_EXTENSION pDevExt = DeferredContext;

	ASSERT(KeGetCurrentIrql() <= DISPATCH_LEVEL);

//	DBGOUT(("Ctg_ClrRtsDpc occured on timer!"));

//	Ctg_SetParityBit(pDevExt); // to save time within the next transmission
	// that was a good idea but we have to wait settling RTS anyway.
//	pDevExt->bParity = TRUE;
/*
	if(pDevExt == NULL)
	{
		DBGOUT(("Ctg_ClrRtsDpc: pDevExt == NULL!!!"));
	}

	if(pDevExt->pTxPktCurr == NULL)
	{
		DBGOUT(("Ctg_ClrRtsDpc: pDevExt->pTxPktCurr == NULL"));
	}
*/
	// set one stop bit for successful data receiving
//		DBGOUT(("Ctg_ClrRtsDpc: before Ctg_SetOneStopBit"));
	Ctg_SetOneStopBit(pDevExt);
//		DBGOUT(("Ctg_ClrRtsDpc: before Ctg_ClrRTS"));

	Ctg_ClrRTS(pDevExt);
//		DBGOUT(("Ctg_ClrRtsDpc: after Ctg_ClrRTS"));
	pDevExt->bRTSIsOn = FALSE;
}

// Function  reads bytes immediately!
NTSTATUS ReadPort(PDEVICE_EXTENSION pDevExt)
{
	PIRP Irp;
	KEVENT evIrpComplete;

	NTSTATUS ntStatus = STATUS_SUCCESS;

	ASSERT(KeGetCurrentIrql() <= DISPATCH_LEVEL);

	Irp = IoAllocateIrp(pDevExt->filterDevObj->StackSize, FALSE);

	if(Irp == NULL)
	{
		DBGOUT(("Could not allocate IRP with IoAllocateIrp!\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		PIO_STACK_LOCATION stack = IoGetNextIrpStackLocation(Irp);

		stack->MajorFunction = IRP_MJ_READ;
		stack->Parameters.Read.Length  = 10; // receive max 10 bytes 
		Irp->AssociatedIrp.SystemBuffer = pDevExt->pRxBufferRaw;

		KeInitializeEvent( &evIrpComplete, SynchronizationEvent, FALSE );

		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)OnEmptyIrpComplete, (PVOID) pDevExt, TRUE, TRUE, TRUE);
		
		ntStatus = IoCallDriver(pDevExt->topDevObj, Irp);
		
		if (!NT_SUCCESS(ntStatus))
		{
			DBGOUT(("ReadPort: Error using IoCallDriver: ntStatus=%x (%d)!\n", ntStatus, ntStatus));
		}
		else // NT_SUCCESS = TRUE
		{
			PUCHAR pByte;
			ULONG i;

			if(ntStatus == STATUS_PENDING)
			{
				// Wait for IRP completion
				DBGOUT(("ReadPort: Waiting KeWaitForSingleObject\n", ntStatus, ntStatus));

				KeWaitForSingleObject( &evIrpComplete, Executive, KernelMode, FALSE, NULL );
			}

			pByte = Irp->AssociatedIrp.SystemBuffer;
			
			for(i=0; i<Irp->IoStatus.Information; i++)
			{
//				DBGOUT(("ReadPort: byte received = %x (%c).\n", *pByte, *pByte));
				pDevExt->pRxBufferRaw[pDevExt->RxBufferRawWriteIndex++] = *pByte++;
				if(pDevExt->RxBufferRawWriteIndex >= RX_BUFFER_RAW_SIZE)
					pDevExt->RxBufferRawWriteIndex = 0;
			}
			IoFreeIrp(Irp);
		}
	}

	return ntStatus;
}

NTSTATUS WritePort(PDEVICE_EXTENSION pDevExt)
{
	#define pPktSend (pDevExt->pTxPktCurr)

	PIRP Irp;
	NTSTATUS ntStatus = STATUS_SUCCESS;
	BOOLEAN bParity = FALSE;

	ASSERT(KeGetCurrentIrql() <= DISPATCH_LEVEL);

	// Do we have bytes to write?
/*
	if(pPktSend == NULL)
	{
		DBGOUT(("WritePort entered with pPktSend=NULL"));
	}
	
	if(pDevExt == NULL)
	{
		DBGOUT(("WritePort entered with pDevExt=NULL"));
	}
*/
	// if we have wrong RTS now change it to a valid value
	if(pDevExt->bRTSIsOn == TRUE && pPktSend == NULL)
	{

//		DBGOUT(("WritePort: pDevExt->bRTSIsOn == TRUE && pPktSend == NULL"));
		// We must set a timer because otherwise RTS is cleared immediately(before data are sent).
		KeSetTimer(&pDevExt->ClrRtsTimer, 
			RtlConvertLongToLargeInteger(-(LONG)(10000)), // 1000 us is more or less reliable
			&pDevExt->ClrRtsDpc);

//		Ctg_ClrRTS(pDevExt);
//		pDevExt->bRTSIsOn = FALSE;
	}

	if(pDevExt->bRTSIsOn == FALSE && pPktSend != NULL)
	{
		LONG delay;

		Ctg_SetRTS(pDevExt);
		pDevExt->bRTSIsOn = TRUE;
	// After raising RTS we MUST wait a bit to allow RS-485 converter settling. 
	// During this time, let's set parity (we always send first byte with mark parity) and 
	// then wait a little.
		Ctg_SetParityBit(pDevExt);
		pDevExt->bParity = TRUE;

/*
		// We must set a timer because otherwise RTS is raised just before transmitting.
		KeSetTimer(&pDevExt->ClrRtsTimer, 
			RtlConvertLongToLargeInteger(-(LONG)(4000)), // 1000 us is more or less reliable
			&pDevExt->ClrRtsDpc);
		return ntStatus;
		*/

		// KeStallExecutionProcessor waits for not more than 50 us!
		// If we use timer here we get too long interval and CTG bus does not allow it. So
		// very rarely (several times per second) we can wait here.
		delay = pDevExt->driver_parameters.DelayAfterRtsSet;
		if(delay >= 500) // use KeDelayExecutionThread() only at IRQL = PASSIVE_LEVEL
		{
			if(KeGetCurrentIrql() == PASSIVE_LEVEL)
			{
				LARGE_INTEGER Interval;
				Interval.LowPart = 10*delay;
				Interval.HighPart = 0;
				if(KeDelayExecutionThread(KernelMode, FALSE, &Interval) != STATUS_SUCCESS)
					delay=499; // Flag that we should attempt to wait below since this wait failed
			}
			else delay=499; // cannot KeDelayExecutionThread at IRQL > PASSIVE_LEVEL
		}
		
		// wait here if delay is small or if the previous condition/wait failed
		if(delay < 500)
		{
			while(delay>0)
			{
				KeStallExecutionProcessor(40); // wait for 40 us more
				delay -= 40;
			}
		}

	}

	// if nothing to send we'll just change RTS.
	if(pPktSend == NULL)
	{
//		DBGOUT(("WritePort: pPktSend == NULL, returning"));
		return ntStatus;
	}

	// if we have sent the first byte (with 9 bit set) we will wait until it is completely sent
	// before clearing 9th bit
	if(pDevExt->bFirstByteJustSent)
	{
		LONG delay;
		pDevExt->bFirstByteJustSent = FALSE;
		// We must wait because otherwise 9 bit is cleared before 1st byte is sent (serial.sys bug?)
/*
		KeSetTimer(&pDevExt->ClrRtsTimer, 
			RtlConvertLongToLargeInteger(-(LONG)(4000)), // 1000 us is more or less reliable
			&pDevExt->ClrRtsDpc);
		return ntStatus;
		*/
		
		// KeStallExecutionProcessor waits for not more than 50 us!
		// If we use timer here we get too long interval and CTG bus does not allow it. So
		// very rarely (several times per second) we can wait here.

		delay = pDevExt->driver_parameters.DelayAfterFirstByte;
		
		if(delay >= 500) // use KeDelayExecutionThread() only at IRQL = PASSIVE_LEVEL
		{
			if(KeGetCurrentIrql() == PASSIVE_LEVEL)
			{
				LARGE_INTEGER Interval;
				Interval.LowPart = 10*delay;
				Interval.HighPart = 0;
				if(KeDelayExecutionThread(KernelMode, FALSE, &Interval) != STATUS_SUCCESS)
					delay=499; // Flag that we should attempt to wait below since this wait failed
			}
			else delay=499; // cannot KeDelayExecutionThread at IRQL > PASSIVE_LEVEL
		}
		
		// wait here if delay is small or if the previous condition/wait failed
		if(delay < 500)
		{
			while(delay>0)
			{
				KeStallExecutionProcessor(40); // wait for 40 us more
				delay -= 40;
			}
		}

	}

	// send first byte in packet with parity mark, the rest - with parity spaced
	bParity = (pPktSend->Position)?FALSE:TRUE;

	// set proper parity if needed
	if(pDevExt->bParity != bParity)
	{
		if(bParity == TRUE)
			Ctg_SetParityBit(pDevExt);
		else
			Ctg_ClrParityBit(pDevExt);
		pDevExt->bParity = bParity;
	}

	Irp = IoAllocateIrp(pDevExt->filterDevObj->StackSize, FALSE);

	if(Irp == NULL)
	{
		DBGOUT(("WritePort: Could not allocate IRP with IoAllocateIrp!\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		PIO_STACK_LOCATION stack = IoGetNextIrpStackLocation(Irp);

		stack->MajorFunction = IRP_MJ_WRITE;
		stack->Parameters.Write.Length  = 1;
		Irp->AssociatedIrp.SystemBuffer = pPktSend->data+pPktSend->Position;

		ASSERT(pPktSend != NULL);

		// set flag for waiting (we just sent first byte)
		if(pPktSend->Position == 0)
			pDevExt->bFirstByteJustSent = TRUE;
		
//		DBGOUT(("WritePort: sending %x (%d)", pPktSend->data[pPktSend->Position], 
//			pPktSend->data[pPktSend->Position]));
		
		pPktSend->Position++;
		if(pPktSend->Position >= pPktSend->dataLength)
		{
//			DBGOUT(("WritePort: set pDevExt->pTxPktCurr = NULL"));
			/*pDevExt->pTxPktCurr = */pPktSend = NULL; // all bytes are sent
		}

//		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)OnEmptyIrpCompleteFree, (PVOID) pDevExt, TRUE, TRUE, TRUE);
		IoSetCompletionRoutine(Irp, (PIO_COMPLETION_ROUTINE)OnWriteIrpComplete, (PVOID) pDevExt, TRUE, TRUE, TRUE);
		
//		DBGOUT(("WritePort: Before IoCallDriver\n"));

		ntStatus = IoCallDriver(pDevExt->topDevObj, Irp);
		
//		DBGOUT(("ReadPort: IoCallDriver returns %d (%s)\n", ntStatus, 
//			(ntStatus==STATUS_PENDING)?"STATUS_PENDING":((ntStatus==STATUS_SUCCESS)?"STATUS_SUCCESS":"Unknown!"))); 
		
		if (!NT_SUCCESS(ntStatus))
		{
			DBGOUT(("WritePort: Error using IoCallDriver: ntStatus=%x (%d)!\n", ntStatus, ntStatus));
		}
		else 
		{
			if(ntStatus != STATUS_PENDING)// if NT_SUCCESS 
			{
				DBGOUT(("WritePort: ntStatus != STATUS_PENDING, freeing IRP immediately.\n", ntStatus, ntStatus));
	//			IoFreeIrp(Irp); IRP allocated by IoAllocateIrp should be freed ___from within OnStartWaitMaskIrpComplete!!!!!!!_____
			}
		}		
	}

	return ntStatus;
#undef pPktSend 
}

// primitives

NTSTATUS Ctg_SetRTS(PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS status = STATUS_SUCCESS;
//	DBGOUT(("Ctg_SetRTS"));
	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_SET_RTS, 0);
	if(!NT_SUCCESS(status))
		DBGOUT(("Ctg_SetRTS FAILED: status = %d", status));
	return status;
}

NTSTATUS Ctg_ClrRTS(PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS status = STATUS_SUCCESS;
//	DBGOUT(("Ctg_ClrRTS"));
	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_CLR_RTS, 0);
	if(!NT_SUCCESS(status))
		DBGOUT(("Ctg_ClrRTS FAILED: status = %d", status));
	return status;
}


// To provide slower comamnds transmitting we will try to use 2 stop bits while 
// transmitting and 1 stop bit while receiving.
NTSTATUS Ctg_SetOneStopBit(PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS status = STATUS_SUCCESS;
	PSERIAL_LINE_CONTROL pCtl = (PSERIAL_LINE_CONTROL)pDevExt->pIOCtlBuffer;

	pCtl->Parity = MARKPARITY;
	pCtl->StopBits=ONESTOPBIT;
	pCtl->WordLength=8;

	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_SET_LINE_CONTROL, sizeof(SERIAL_LINE_CONTROL));
	
	if(!NT_SUCCESS(status))
		DBGOUT(("Ctg_SetParityBit FAILED: status = %d", status));
	
	return status;
}

NTSTATUS Ctg_SetParityBit(PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS status = STATUS_SUCCESS;
	PSERIAL_LINE_CONTROL pCtl = (PSERIAL_LINE_CONTROL)pDevExt->pIOCtlBuffer;

//	DBGOUT(("Ctg_SetParityBit"));

/*	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_GET_LINE_CONTROL, sizeof(SERIAL_LINE_CONTROL));
	
	if(!NT_SUCCESS(status))
	{
		DBGOUT(("Ctg_SetParityBit (IOCTL_SERIAL_GET_LINE_CONTROL) FAILED: status = %d", status));
	}
	else
	{
		DBGOUT(("Ctg_ClrParityBit: pCtl->StopBits=%d, pCtl->Parity=%d, pCtl->WordLength=%d", 
			pCtl->StopBits, pCtl->Parity, pCtl->WordLength));
	}
*/

	//pCtl->StopBits=0, pCtl->Parity=*, pCtl->WordLength=8	
	pCtl->Parity = MARKPARITY;
	pCtl->StopBits=TWOSTOPBITS;//ONESTOPBIT;
	pCtl->WordLength=8;

	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_SET_LINE_CONTROL, sizeof(SERIAL_LINE_CONTROL));
	
	if(!NT_SUCCESS(status))
		DBGOUT(("Ctg_SetParityBit FAILED: status = %d", status));
	
	return status;
}

NTSTATUS Ctg_ClrParityBit(PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS status = STATUS_SUCCESS;
	PSERIAL_LINE_CONTROL pCtl = (PSERIAL_LINE_CONTROL)pDevExt->pIOCtlBuffer;

//	DBGOUT(("Ctg_ClrParityBit"));

/*	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_GET_LINE_CONTROL, sizeof(SERIAL_LINE_CONTROL));
	
	if(!NT_SUCCESS(status))
	{
		DBGOUT(("Ctg_ClrParityBit (IOCTL_SERIAL_GET_LINE_CONTROL) FAILED: status = %d", status));
	}
	else
	{
		DBGOUT(("Ctg_ClrParityBit: pCtl->StopBits=%d, pCtl->Parity=%d, pCtl->WordLength=%d", 
			pCtl->StopBits, pCtl->Parity, pCtl->WordLength));
	}
*/
	pCtl->Parity = SPACEPARITY;
	pCtl->StopBits=TWOSTOPBITS;//ONESTOPBIT;
	pCtl->WordLength=8;
	
	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_SET_LINE_CONTROL, sizeof(SERIAL_LINE_CONTROL));
	if(!NT_SUCCESS(status))
		DBGOUT(("Ctg_ClrParityBit FAILED: status = %d", status));
	return status;
}

NTSTATUS Ctg_PurgeRx(PDEVICE_EXTENSION pDevExt)
{
	NTSTATUS status = STATUS_SUCCESS;
	PULONG pMask = (PULONG)pDevExt->pIOCtlBuffer;
	*pMask = SERIAL_PURGE_RXCLEAR;
	
	status = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_PURGE, sizeof(ULONG));

	if(!NT_SUCCESS(status))
	{
		DBGOUT(("Ctg_ClrRTS FAILED: status = %d", status));
	}
	else
	{
//		DBGOUT(("Ctg_PurgeRx"));
	}

	return status;
}



NTSTATUS Ctg_InitStorage(PDEVICE_EXTENSION pDevExt)
{
	RtlZeroMemory(pDevExt->pRxBufferRaw, sizeof(pDevExt->pRxBufferRaw)); 

	pDevExt->RxBufferRawWriteIndex = pDevExt->RxBufferRawReadIndex = 0;

	RtlZeroMemory(pDevExt->arRxPkt, sizeof(pDevExt->arRxPkt)); 
	pDevExt->NextRxPktRead = pDevExt->NextRxPktWrite = 0; 
	pDevExt->pRxPktCurr = NULL;

	RtlZeroMemory(pDevExt->arTxPkt, sizeof(pDevExt->arTxPkt)); 
	pDevExt->NextTxPktRead = pDevExt->NextTxPktWrite = 0; 
	pDevExt->pTxPktCurr = NULL;

	// port status bits
	pDevExt->bRTSIsOn = FALSE; // is RTS on?

	RtlZeroMemory(&pDevExt->pktACKPacket, sizeof(pDevExt->pktACKPacket)); 

	// initialize our private data
	pDevExt->bCtgDriverIsOn = FALSE; // filter is initially silent
	pDevExt->pPendingIrp = NULL; 
	pDevExt->RxBufferRawReadIndex = pDevExt->RxBufferRawWriteIndex = 0;

	pDevExt->RVUAddress = 123; // set initial address
	Ctg_InitACKPacket(pDevExt, FALSE); // AFTER THAT init ACK packet

	pDevExt->pRxPktCurr = pDevExt->arRxPkt; // first packet is filled
	pDevExt->pTxPktCurr = NULL; // nothing is sent

	pDevExt->diag_info.nMeaningfulPacketsReceived = 0; // nimber of non-poll packets received
	pDevExt->diag_info.nMeaningfulPacketsTransmitted = 0; // number of non-ACK packets sent
	pDevExt->diag_info.nTotalPacketsReceived = 0; // nimber of poll packets received
	pDevExt->diag_info.nTotalPacketsTransmitted = 0; // number of packets sent
	pDevExt->diag_info.TimeoutSinceLastPacketReceived = 3; // number of timer ticks
	
	pDevExt->driver_parameters.DelayAfterFirstByte = 320; // default values for Futura
	pDevExt->driver_parameters.DelayAfterRtsSet = 280; // default values for Futura
	pDevExt->driver_parameters.bTestMode = FALSE; // no emulation

//	DBGOUT(("sizeof(PACKET)=%d, sizeof(PPACKET)=%d",
//		sizeof(PACKET), sizeof(PPACKET)));

	return STATUS_SUCCESS;
}

VOID Ctg_InitACKPacket(IN PDEVICE_EXTENSION pDevExt, IN BOOLEAN bTestMode)
{
	UCHAR ACKCmd[5]={1, 123, 1, 12, 226};
	UCHAR TestCmd[6]={1, 123, 2, 85, 1, 45};
	PUCHAR pCmd;
	UCHAR diff;
	ULONG size = (bTestMode==TRUE)?sizeof(TestCmd):sizeof(ACKCmd);
	pCmd = (bTestMode==TRUE)?TestCmd:ACKCmd;

	if(pDevExt->RVUAddress < 122 || pDevExt->RVUAddress > 124)
		pDevExt->RVUAddress = 123;

	// count difference between actual address and initial one
	diff = pDevExt->RVUAddress - pCmd[1]; 
	pCmd[1] = pDevExt->RVUAddress;
	pCmd[4] += diff; // adjust checksum

	// ACK packet: 1, 123, 1, 12, 226
	RtlMoveMemory(pDevExt->pktACKPacket.data, pCmd, size);

	pDevExt->pktACKPacket.bPacketValid = TRUE;
	pDevExt->pktACKPacket.dataLength = size;
	pDevExt->pktACKPacket.Length = size-4;
	pDevExt->pktACKPacket.Position = 0;
}

// set mask to wait (from this moment we are waiting for new event)
NTSTATUS Ctg_SetWaitMask(PDEVICE_EXTENSION pDevExt, ULONG mask)
{
	NTSTATUS ntStatus = STATUS_SUCCESS;
	pDevExt->pIOCtlBuffer[0]=(UCHAR)mask;
	pDevExt->pIOCtlBuffer[1]=0;
	pDevExt->pIOCtlBuffer[2]=0;
	pDevExt->pIOCtlBuffer[3]=0;

	ntStatus = SendIoctlSerialImmediate(pDevExt, IOCTL_SERIAL_SET_WAIT_MASK, sizeof(ULONG));
	if(!NT_SUCCESS(ntStatus))
	{
		DBGOUT(("Error %x (%d) calling SendIoctlSerialImmediate!\n", ntStatus, ntStatus));
	}

	return ntStatus;
}

NTSTATUS Ctg_ClearPacket(PPACKET pPkt)
{
	ASSERT(pPkt);

	return STATUS_SUCCESS;
	if(pPkt)
	{
//		RtlZeroMemory(pPkt, sizeof(_PACKET));
		pPkt->bPacketValid = FALSE;
		pPkt->Position = 0; 
		pPkt->Length = 0;
		pPkt->dataLength = 0;
		pPkt->Checksum = 0;
		pPkt->bPacketStarted = FALSE;
	}
	
	return STATUS_SUCCESS;
}
